''' Progress bar utilities
'''

from typing import Dict, Any

import tqdm
from pyckmeans.core import MultiCKMeans

class MultiCKMeansProgressBars:
    '''MultiCKMeansProgressBars

    Context Manager for a MultiCKMeans progress bars.

    Parameters
    ----------
    mckm : MultiCKMeans
        MultiCKMeans object to display progress bars for.
    kwargs : Dict[str, Any]
        Additional keyword arguments passed to tqdm.tqdm.
    '''
    def __init__(
        self,
        mckm: MultiCKMeans,
        **kwargs: Dict[str, Any],
    ):
        self.mckm = mckm

        self.ks = mckm.k
        self.n_rep = mckm.n_rep

        self._ckm_idx = 0
        self._iter = 0
        self._done = False

        # tqdm options
        self._tqdm_kwargs = {
        }
        self._tqdm_kwargs.update(kwargs)

        # init first progress bar
        self._tqdm = tqdm.tqdm(
            total=self.n_rep,
            mininterval=0.5,
            desc=f'k={self.ks[self._ckm_idx]}',
            **self._tqdm_kwargs,
        )

    def update(
        self,
        n: int = 1,
    ):
        '''update

        Update progress by n iterations.

        Parameters
        ----------
        n : int, optional
            Progress increment in iterations, by default 1
        '''
        if self._done:
            return

        self._iter += n
        self._tqdm.update(n)

        if self._iter >= self.n_rep:
            self._tqdm.close()
            self._iter = 0
            self._ckm_idx += 1
            if self._ckm_idx >= len(self.ks):
                self._done = True
            else:
                self._tqdm = tqdm.tqdm(
                    total=self.n_rep,
                    desc=f'k={self.ks[self._ckm_idx]}',
                    **self._tqdm_kwargs,
                )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._tqdm.close()
        return



'''multickmeans module'''

from typing import List, Optional, Iterable, Dict, Any, Tuple, Union, Callable, TYPE_CHECKING
import numpy
import pandas

from pyckmeans.ordination import PCOAResult
from .ckmeans import CKmeansResult, CKmeans, InvalidClusteringMetric

if TYPE_CHECKING:
    import matplotlib
    import matplotlib.figure

class MultiCKmeansResult:
    '''MultiCKmeansResult

    Result of MultiCKmeansResult.predict.

    Parameters
    ----------
    ckmeans_results: List[CKmeansResult]
        List of CKmeansResults.
    names: Optional[Iterable(str)]
        Sample names.
    '''
    def __init__(
        self,
        ckmeans_results: List[CKmeansResult],
        names: Optional[Iterable[str]] = None,
    ):
        self.ckmeans_results = ckmeans_results
        self.names = numpy.arange(ckmeans_results[0].cmatrix.shape[0]).astype(str) \
            if names is None else numpy.array(names).astype(str)

        self.ks = numpy.array([ckm_res.k for ckm_res in ckmeans_results])

        self.sils = numpy.array([ckm_res.sil for ckm_res in ckmeans_results])
        self.bics = numpy.array([ckm_res.bic for ckm_res in ckmeans_results])
        self.dbs = numpy.array([ckm_res.db for ckm_res in ckmeans_results])
        self.chs = numpy.array([ckm_res.ch for ckm_res in ckmeans_results])

        self.metrics = pandas.DataFrame({
            'k': self.ks,
            'sil': self.sils,
            'bic': self.bics,
            'db': self.dbs,
            'ch': self.chs,
        })

    def order(
        self,
        by: int,
        method: str = 'GW',
        linkage_type: str = 'average',
    ) -> numpy.ndarray:
        '''order

        Get optimal sample order according to hierarchical clustering of the
        CKmeansResult at index "by".

        Parameters
        ----------
        by : int
            Index of the CKMeansResult to order by.
        method : str
            Reordering method. Either 'GW' (Gruvaeus & Wainer, 1972) or 'OLO' for
            scipy.hierarchy.optimal_leaf_ordering.

            Gruvaeus, G., H., Wainer. 1972. Two Additions to Hierarchical Cluster Analysis.
            The British Psychological Society 25.
        linkage_type : str
            Linkage type for the hierarchical clustering. One of

            * 'average'
            * 'complete'
            * 'single'
            * 'weighted'
            * 'centroid'

            See scipy.cluster.hierarchy.linkage for details.

        Returns
        -------
        numpy.ndarray
            Optimal sample order.
        '''
        ckm_res = self.ckmeans_results[by]

        return ckm_res.order(method=method, linkage_type=linkage_type)

    def sort(
        self,
        by: int,
        method: str = 'GW',
        linkage_type: str = 'average',
        in_place: bool = False,
    ) -> 'MultiCKmeansResult':
        '''sort

        Sort samples according to hierarchical clustering of the
        CKmeansResult at index "by".

        Parameters
        ----------
        by : int
            Index of the CKMeansResult to sort by.
        method : str
            Reordering method. Either 'GW' (Gruvaeus & Wainer, 1972) or 'OLO' for
            scipy.hierarchy.optimal_leaf_ordering.

            Gruvaeus, G., H., Wainer. 1972. Two Additions to Hierarchical Cluster Analysis.
            The British Psychological Society 25.
        linkage_type : str
            Linkage type for the hierarchical clustering. One of

            * 'average'
            * 'complete'
            * 'single'
            * 'weighted'
            * 'centroid'

            See scipy.cluster.hierarchy.linkage for details.
        in_place : bool
            If False, a new, sorted MultiCKmeansResult object will be returned.
            If True, the object will be sorted in place and self will be returned.

        Returns
        -------
        MultiCKmeansResult
            Sorted MultiCKmeansResult
        '''

        order = self.order(by=by, method=method, linkage_type=linkage_type)

        return self.reorder(order, in_place=in_place)

    def reorder(
        self,
        order: numpy.ndarray,
        in_place: bool = False,
    ) -> 'MultiCKmeansResult':
        '''reorder

        Reorder samples in all CKmeansResults according to provided order.

        Parameters
        ----------
        order : numpy.ndarray
            New sample order.
        in_place : bool
            If False, a new, sorted MultiCKmeansResult object will be returned.
            If True, the object will be sorted in place and self will be returned.

        Returns
        -------
        MultiCKmeansResult
            Reordered MultiCKmeansResult
        '''

        if in_place:
            mckmres = self
            for ckmres in self.ckmeans_results:
                ckmres.reorder(order, in_place=True)
        else:
            ckm_results = [
                ckmres.reorder(order, in_place=False) for ckmres in self.ckmeans_results
            ]
            names = None if self.names is None else self.names.copy()
            mckmres = MultiCKmeansResult(ckm_results, names=names)

        return mckmres

    def plot_metrics(
        self,
        figsize: Tuple[float, float] = (7, 7),
    ) -> 'matplotlib.figure.Figure':
        '''plot_metrics

        Plot MultiCKMeansResult metrics.

        Parameters
        ----------
        figsize : Tuple[float, float], optional
            Figure size for the matplotlib figure, by default (7, 7).

        Returns
        -------
        matplotlib.figure.Figure
            Matplotlib Figure of the metrics plot.
        '''

        from pyckmeans.utils import plot_multickmeans_metrics

        fig = plot_multickmeans_metrics(
            mckm_res=self,
            figsize=figsize,
        )
        fig.tight_layout()
        return fig

class MultiCKMeans:
    '''MultiCKMeans

    Convenience class wrapping Consensus K-Means runs for multiple different numbers of clusters.

    Parameters
    ----------
    k : Iterable[int]
        List of cluster counts for CKmeans.
    n_rep : int, optional
        Number of K-Means to fit for each single CKmeans, by default 100
    p_samp : float, optional
        Proportion of samples (observations) to randomly draw per K-Means run, by default 0.8.
        The resulting number of samples will be rounded up. I.e. if number of samples is 10 and
        p_samp is 0.75, each K-Means will use 8 randomly drawn samples (0.72 * 10 = 7.2, 7.2 -> 8).
    p_feat : float, optional
        Proportion of features (predictors) to randomly draw per K-Means run, by default 0.8.
        The resulting number of features will be rounded up. I.e. if number of features is 10 and
        p_feat is 0.72, each K-Means will use 8 randomly drawn features (0.72 * 10 = 7.5, 7.2 -> 8).
    metrics : Iterable[str]
        Clustering quality metrics to calculate while training. Available metrics are
        * "sil" (Silhouette Index)
        * "bic" (Bayesian Information Criterion)
        * "db" (Davies-Bouldin Index)
        * "ch" (Calinski-Harabasz).
    kwargs : Dict[str, Any]
        Additional keyword arguments passed to sklearn.cluster.KMeans.
    '''
    def __init__(
        self,
        k: Iterable[int],
        n_rep: int = 100,
        p_samp: float = 0.8,
        p_feat: float = 0.8,
        metrics: Iterable[str] = ('sil', 'bic'),
        **kwargs: Dict[str, Any],
    ):
        self.k = k
        self.n_rep = n_rep
        self.p_samp = p_samp
        self.p_feat = p_feat

        for metric in metrics:
            if not metric in CKmeans.AVAILABLE_METRICS:
                am_str = ", ".join(CKmeans.AVAILABLE_METRICS)
                msg = f'Unknown metric "{metric}". Available metrics are {am_str}.'
                raise InvalidClusteringMetric(msg)

        self._metrics = metrics
        self._kmeans_kwargs = kwargs

        self.ckmeans: Optional[List[CKmeans]] = None

    def fit(
        self,
        x: Union[numpy.ndarray, PCOAResult, pandas.DataFrame],
        progress_callback: Optional[Callable] = None,
    ):
        '''fit

        Fit MultiCKmeans.

        Parameters
        ----------
        x : Union[numpy.ndarray, PCOAResult]
            a n * m matrix (numpy.ndarray) or dataframe (pandas.DataFrame), where n is the number
            of samples (observations) and m is the number of features (predictors).
            Alternatively a pyckmeans.ordination.PCOAResult as returned from pyckmeans.pcoa.
        progress_callback : Optional[Callable]
            Optional callback function for progress reporting.
        '''

        if isinstance(x, PCOAResult):
            x = x.vectors
        elif isinstance(x, pandas.DataFrame):
            x = x.values

        # _fit is called here to be able to extend later on.
        # The plan is to add a parallel fitting function later on
        # e.g. _fit_parallel(x, progress_callback, n_cores)
        self._fit(x, progress_callback=progress_callback)

    def predict(
        self,
        x: Union[numpy.ndarray, PCOAResult, pandas.DataFrame],
        linkage_type: str = 'average',
        return_cls: bool = False,
        progress_callback: Optional[Callable] = None,
    ) -> MultiCKmeansResult:
        '''predict

        Predict cluster membership of new data from all fitted CKmeans.

        Parameters
        ----------
        x : Union[numpy.ndarray, PCOAResult]
            a n * m matrix (numpy.ndarray) or dataframe (pandas.DataFrame), where n is the number
            of samples (observations) and m is the number of features (predictors). If x is a
            dataframe, the index will be used a sample names.
            Alternatively a pyckmeans.ordination.PCOAResult as returned from pyckmeans.pcoa.
        linkage_type : str
            Linkage type of the hierarchical clustering that is used for consensus cluster
            calculation. One of

            * 'average'
            * 'complete'
            * 'single'
            * 'weighted'
            * 'centroid'

            See scipy.cluster.hierarchy.linkage for details.
        return_cls : bool
            If True, the cluster memberships of the single K-Means runs will be present
            in the output.
        progress_callback : Optional[Callable]
            Optional callback function for progress reporting.

        Returns
        -------
        CKmeansResult
            Object comprising a  n * n consensus matrix, and a n-length vector of
            precited cluster memberships.
        '''
        names = None
        if isinstance(x, PCOAResult):
            names = numpy.array(x.names).astype(str)
        elif isinstance(x, pandas.DataFrame):
            names = numpy.array(x.index).astype(str)

        ckmeans_results: List[CKmeansResult] = []
        for ckm in self.ckmeans:
            ckm_res = ckm.predict(
                x=x,
                linkage_type=linkage_type,
                return_cls=return_cls,
                progress_callback=progress_callback,
            )
            ckmeans_results.append(ckm_res)

        return MultiCKmeansResult(
            ckmeans_results=ckmeans_results,
            names=names,
        )

    def _fit(
        self,
        x: numpy.ndarray,
        progress_callback: Optional[Callable] = None,
    ):
        '''_fit

        Internal sequential fitting function.

        Parameters
        ----------
        x : numpy.ndarray
            n * m matrix, where n is the number of samples (observations) and m is
            the number of features (predictors).
        progress_callback : Optional[Callable]
            Optional callback function for progress reporting.
        '''
        self._reset()

        self.ckmeans = []

        for k in self.k:
            ckm = CKmeans(
                k=k,
                n_rep=self.n_rep,
                p_samp=self.p_samp,
                p_feat=self.p_feat,
                metrics=self._metrics,
                **self._kmeans_kwargs,
            )
            ckm.fit(x=x, progress_callback=progress_callback)
            self.ckmeans.append(ckm)

    def _reset(self):
        '''_reset

        Reset MultiCKmeans object.
        '''
        self.ckmeans = None



''' csv

    Comma Separated Value (CSV) input and output.
'''
import os
from typing import Optional

import pandas

import pyckmeans.distance

class InvalidMatrixShapeError(Exception):
    '''InvalidMatrixShapeError'''

class IncompatibleNamesError(Exception):
    '''IncompatibleNamesError'''

def read_csv_distmat( # pylint: disable=missing-param-doc
    file_path: str,
    header: Optional[int] = 0,
    index_col: Optional[int] = 0,
    sep: str = ',',
    **kwargs,
) -> 'pyckmeans.distance.DistanceMatrix':
    '''read_csv_distmat

    Read distance matrix from CSV file.

    Parameters
    ----------
    file_path : str
        Path to CSV file.
    header : Optional[int]
        Determines the row in the CSV file containing
        sample names. Is passed to pandas.read_csv(). By default 0, meaning
        the first row.
    index_col : Optional[int]
        Determines the index column. By default, the first column is expected
        to contain sample names. Passed to pandas.read_csv().
    sep : str
        Column separator, be default ','. Passed to Passed to pandas.read_csv().
    **kwargs
        Additional keyword arguments passed to pandas.read_csv().
    Returns
    -------
    pyckmeans.distance.DistanceMatrix
        DistanceMatrix object.

    Raises
    ------
    InvalidMatrixShapeError
        Raised if matrix is not square.
    IncompatibleNamesError
        Raised if column and row names do not match.
    '''
    dist_df = pandas.read_csv(
        file_path,
        header=header,
        index_col=index_col,
        sep=sep,
        **kwargs
    )

    dist_mat = dist_df.values

    # distance matrix must be a square matrix
    if dist_mat.shape[0] != dist_mat.shape[1]:
        msg = 'Expected a square matrix but matrix has dimensions '+ \
            f'{dist_mat.shape[0]}x{dist_mat.shape[1]}.'
        raise InvalidMatrixShapeError(msg)

    names = None
    # names are present in file
    if (not header is None) or (not index_col is None):
        # row and column names are present
        if (not header is None) and (not index_col is None):
            names_a = [nm.strip() for nm in dist_df.index.astype(str)]
            names_b = [nm.strip() for nm in dist_df.columns.astype(str)]

            # if row names and column names do not match, something
            # is probably wrong
            if not all([a == b for a, b in zip(names_a, names_b)]):
                raise IncompatibleNamesError('Column and row names do not match.')

            names = names_a
        # column names are present
        elif not header is None:
            names = [nm.strip() for nm in dist_df.columns.astype(str)]
        # row names are present
        elif not index_col is None:
            names = [nm.strip() for nm in dist_df.index.astype(str)]


    return pyckmeans.distance.DistanceMatrix(
        dist_mat,
        names,
    )

def write_csv_distmat(
    dist: 'pyckmeans.distance.DistanceMatrix',
    file_path: str,
    force: bool = False,
) -> None:
    '''write_csv_distmat

    Write DistanceMatrix object to CSV.

    Parameters
    ----------
    dist : pyckmeans.distance.DistanceMatrix
        DistanceMatrix object.
    file_path : str
        CSV file path.
    force : bool, optional
        Force overwrite if file_path already exists, by default False

    Raises
    ------
    FileExistsError
        Raised if file at file_path already exists and force is False.
    FileExistsError
        Raised if file_path points to an existing directory.
    '''
    if os.path.exists(file_path):
        if os.path.isfile(file_path) and not force:
            msg = f'File {file_path} already exists. If you want to overwrite ' +\
                'it run the function with force=True.'
            raise FileExistsError(msg)
        else:
            msg = f'A directory exists at path {file_path}.'
            raise FileExistsError(msg)

    dist_df = pandas.DataFrame(
        dist.dist_mat,
        columns=dist.names,
        index=dist.names,
    )

    dist_df.to_csv(file_path, index_label='sample')



''' Module for distance matrix ordering.
'''

from typing import Union
import numpy
from scipy.cluster import hierarchy

import pyckmeans.distance

class InvalidReorderMethod(Exception):
    '''InvalidReorderMethod'''
class InvalidLinkageType(Exception):
    '''InvalidLinkageType'''

REORDER_METHODS = (
    'GW',
    'OLO',
)
LINKAGE_TYPES = (
    'average',
    'complete',
    'single',
    'weighted',
    'centroid',
)


def distance_order(
    dist: Union[numpy.ndarray, 'pyckmeans.distance.DistanceMatrix'],
    method: str = 'GW',
    linkage_type: str = 'average',
) -> numpy.ndarray:
    '''distance_order

    Get optimal distance matrix order.

    Parameters
    ----------
    dist : Union[numpy.ndarray, 'pyckmeans.distance.DistanceMatrix']
        A n * n distance matrix as either numpy.ndarray or as
        pyckmeans.distance.DistanceMatrix object.
    method : str
        Reordering method. Either 'GW' (Gruvaeus & Wainer, 1972) or 'OLO' for
        scipy.hierarchy.optimal_leaf_ordering.

        Gruvaeus, G., H., Wainer. 1972. Two Additions to Hierarchical Cluster Analysis.
        The British Psychological Society 25.
    linkage_type : str
        Linkage type for the hierarchical clustering. One of

        * 'average'
        * 'complete'
        * 'single'
        * 'weighted'
        * 'centroid'

        See scipy.cluster.hierarchy.linkage for details.

    Returns
    -------
    numpy.ndarray
        Optimal order as vector.

    Raises
    ------
    InvalidReorderMethod
        Raised if an unknown reordering method is passed.
    InvalidLinkageType
        Raised if an unknown linakage type is passed.
    '''

    method = method.upper()
    if method not in REORDER_METHODS:
        msg = f'"{method}" is not a valid reordering method. Available ' +\
            f'methods are {REORDER_METHODS}.'
        raise InvalidReorderMethod(msg)

    linkage_type = linkage_type.lower()
    if linkage_type not in LINKAGE_TYPES:
        msg = f'"{linkage_type}" is not a valid linkage type. Available ' +\
            f'types are {LINKAGE_TYPES}.'
        raise InvalidLinkageType(msg)

    is_ndarray = isinstance(dist, numpy.ndarray)
    if is_ndarray:
        dist_mat = dist
    else:
        dist_mat = dist.dist_mat

    dist_mat_cond = condensed_form(dist_mat)
    linkage_mat = hierarchy.linkage(dist_mat_cond, method=linkage_type)
    # cluster distance can become negative due to floating point
    # errors.
    linkage_mat[numpy.abs(linkage_mat) < 1e-8] = 0

    if method == 'OLO':
        linkage_mat = hierarchy.optimal_leaf_ordering(linkage_mat, dist_mat_cond)
    elif method == 'GW':
        linkage_mat = reorder_linkage_gw(linkage_mat, dist_mat)

    order = hierarchy.leaves_list(linkage_mat)
    dist_mat = dist_mat[order, :][:, order]

    return order

# This function duplicates code from distance_order, but I
# want to keep this duplication for now for more flexibility
def reorder_distance(
    dist: Union[numpy.ndarray, 'pyckmeans.distance.DistanceMatrix'],
    method: str = 'GW',
    linkage_type: str = 'average',
) -> Union[numpy.ndarray, 'pyckmeans.distance.DistanceMatrix']:
    '''reorder_distance

    Reorder distance matrix using hierarchical clustering.

    Parameters
    ----------
    dist : Union[numpy.ndarray, 'pyckmeans.distance.DistanceMatrix']
        A n * n distance matrix as either numpy.ndarray or as
        pyckmeans.distance.DistanceMatrix object.
    method : str
        Reordering method. Either 'GW' (Gruvaeus & Wainer, 1972) or 'OLO' for
        scipy.hierarchy.optimal_leaf_ordering.

        Gruvaeus, G., H., Wainer. 1972. Two Additions to Hierarchical Cluster Analysis.
        The British Psychological Society 25.
    linkage_type : str
        Linkage type for the hierarchical clustering. One of

        * 'average'
        * 'complete'
        * 'single'
        * 'weighted'
        * 'centroid'

        See scipy.cluster.hierarchy.linkage for details.

    Returns
    -------
    Union[numpy.ndarray, 'pyckmeans.distance.DistanceMatrix']
        The sorted distance matrix as either numpy.ndarray or
        pyckmeans.distance.DistanceMatrix, depending on the input.

    Raises
    ------
    InvalidReorderMethod
        Raised if an unknown reordering method is passed.
    InvalidLinkageType
        Raised if an unknown linakage type is passed.
    '''
    method = method.upper()
    if method not in REORDER_METHODS:
        msg = f'"{method}" is not a valid reordering method. Available ' +\
            f'methods are {REORDER_METHODS}.'
        raise InvalidReorderMethod(msg)

    linkage_type = linkage_type.lower()
    if linkage_type not in LINKAGE_TYPES:
        msg = f'"{linkage_type}" is not a valid linkage type. Available ' +\
            f'types are {LINKAGE_TYPES}.'
        raise InvalidLinkageType(msg)

    is_ndarray = isinstance(dist, numpy.ndarray)
    if is_ndarray:
        dist_mat = dist
    else:
        dist_mat = dist.dist_mat

    dist_mat_cond = condensed_form(dist_mat)
    linkage_mat = hierarchy.linkage(dist_mat_cond, method=linkage_type)
    if method == 'OLO':
        linkage_mat = hierarchy.optimal_leaf_ordering(linkage_mat, dist_mat_cond)
    elif method == 'GW':
        linkage_mat = reorder_linkage_gw(linkage_mat, dist_mat)

    order = hierarchy.leaves_list(linkage_mat)
    dist_mat = dist_mat[order, :][:, order]

    if is_ndarray:
        return dist_mat
    else:
        return pyckmeans.distance.DistanceMatrix(
            dist_mat,
            dist.names[order] if not dist.names is None else None,
        )

def condensed_form(dist: numpy.ndarray) -> numpy.ndarray:
    '''condensed_form

    Convert n*n distance matrix to condensed vector form.

    Parameters
    ----------
    dist : numpy.ndarray
        n * n distance matrix.

    Returns
    -------
    numpy.ndarray
        Distance matrix in condensed vector form as expected by
        scipy.cluster.hierarchy.linkage.
    '''

    return dist[numpy.triu_indices_from(dist, k=1)]

def reorder_linkage_gw(
    linkage: numpy.ndarray,
    dist: numpy.ndarray,
) -> numpy.ndarray:
    '''reorder_linkage_gw

    Reorder linkage matrix using the algorithm described by Gruvaeus & Wainer (1972) [1]_.

    Parameters
    ----------
    linkage : numpy.ndarray
        Linkage matrix as returned from scipy.cluster.hierarchy.linkage.
    dist : numpy.ndarray
        n * n distance matrix.

    Returns
    -------
    numpy.ndarray
        Reordered linkage matrix.

    References
    ----------
    .. [1]  Gruvaeus, G., H., Wainer. 1972. Two Additions to Hierarchical Cluster Analysis.
            The British Psychological Society 25.
    '''
    linkage = linkage.copy()

    n = linkage.shape[0]

    # left and right leaves of a cluster
    l_r = numpy.zeros((n, 2))
    # matrix determining, whether a cluster (subtree) should be flipped
    flip = numpy.full((n, 2), False)

    # find left and right leaves of clusters
    # and determine, whether cluster should
    # be flipped
    for i in range(n):
        l, r = linkage[i, [0, 1]].astype(int)

        # l and r are singletons
        if l <= n and r <= n:
            l_r[i] = (l, r)
        # only l is a singleton
        elif l <= n:
            l_r[i, 0] = l

            # left and right leaves of cluster r
            rl, rr = l_r[r - (n + 1)].astype(int)

            if dist[l, rl] < dist[l, rr]:
                l_r[i, 1] = rr
            else:
                l_r[i, 1] = rl
                flip[i, 1] = True
        # only r is singleton
        elif r <= n:
            l_r[i, 1] = r

            # left and right leaves of cluster l
            ll, lr = l_r[l - (n + 1)].astype(int)

            if dist[r, ll] < dist[r, lr]:
                l_r[i, 0] = lr
                flip[i, 0] = True
            else:
                l_r[i, 0] = ll
        # none of l and r are singletons
        else:
            # left and right leaves
            ll, lr = l_r[l - (n + 1)].astype(int)
            rl, rr = l_r[r - (n + 1)].astype(int)

            d_ll_rl = dist[ll, rl] # 0
            d_ll_rr = dist[ll, rr] # 1
            d_lr_rl = dist[lr, rl] # 2
            d_lr_rr = dist[lr, rr] # 3

            mn_idx = numpy.argmin([d_ll_rl, d_ll_rr, d_lr_rl, d_lr_rr])
            if mn_idx == 0: # d_ll_rl
                l_r[i] = (lr, rr)
                flip[i, 0] = True
            elif mn_idx == 1: # d_ll_rr
                l_r[i] = (lr, rl)
                flip[i] = (True, True)
            elif mn_idx == 2: # d_lr_rl
                l_r[i] = (ll, rr)
            else: # d_lr_rr
                l_r[i] = (ll, rl)
                flip[i, 1] = True

    # apply flip
    for i in range((n-1), 0, -1):
        if flip[i, 0]:
            c = linkage[i, 0].astype(int)
            # non-singleton cluster
            if c > n:
                c = c - (n + 1)
                linkage[c, [0, 1]] = linkage[c, [1, 0]]
                if flip[c, 0] == flip[c, 1]:
                    flip[c] = ~flip[c]
        if flip[i, 1]:
            c = linkage[i, 1].astype(int)
            if c > n:
                c = c - (n + 1)
                linkage[c, [0, 1]] = linkage[c, [1, 0]]
                if flip[c, 0] == flip[c, 1]:
                    flip[c] = ~flip[c]

    return linkage



''' Knee and elbow search.
'''

from typing import Callable, Iterable
import warnings

import numpy

def rel_extrema_idcs(
    x: numpy.ndarray,
    cmp_fun: Callable[[numpy.ndarray, numpy.ndarray], numpy.ndarray] = numpy.greater,
    mode: str = 'clip',
) -> numpy.ndarray:
    '''rel_extrema_idcs

    Find indices of relative extrema. A relative extremum is found if
    at an element, if cmp_fun returns true for both of its neighbors.

    Parameters
    ----------
    x : numpy.ndarray
        Data vector.
    cmp_fun : Callable[[numpy.ndarray, numpy.ndarray], numpy.ndarray], optional
        Compare function function, by default numpy.greater
    mode : str, optional
        Specifies how out-of-bounds indices will behave.

        * 'raise' - raise an error (default)
        * 'wrap' - wrap around
        * 'clip' - clip to the range

        'clip' mode means that all indices that are too large are replaced
        by the index that addresses the last element along that axis. Note
        that this disables indexing with negative numbers.

        (mode documentation copied from numpy.take)

    Returns
    -------
    numpy.ndarray
        Indices of the extrema.
    '''
    idcs = numpy.arange(0, x.shape[0])
    left = x.take(idcs + 1, mode=mode)
    right = x.take(idcs - 1, mode=mode)

    return numpy.nonzero(cmp_fun(x, left) & cmp_fun(x, right))[0]


# the following code is mostly copied from
# https://github.com/arvkevi/kneed and
# was adapted for compatibility

VALID_CURVE = ('convex', 'concave')
VALID_DIRECTION = ('increasing', 'decreasing')

class KneeLocator:
    '''KneeLocator

    An implementation of the Kneedle algorithm [1]_.

    Once instantiated, this class attempts to find the point of maximum
    curvature on a line. The knee is accessible via the `.knee` attribute.

    Parameters
    ----------
    x : numpy.ndarray
        x values.
    y : numpy.ndarray
        y values.
    S : float, optional
        Sensitivity, original paper suggests default of 1.0, by default 1.0.
    curve : str, optional
        If 'concave', algorithm will detect knees. If 'convex', it
        will detect elbows., by default 'concave'.
    direction : str, optional
        Curve direction. One of {'increasing', 'decreasing'}, by default 'increasing'.
    interp_method : str, optional
        Interpolation method. One of

        * 'interp1d' - no interpolation
        * 'polynomial' - polynomial interpolation

        By default 'interp1d'.
    online : bool, optional
        Correct old knee points if True, will return first knee if False,
        by default False.
    polynomial_degree : int, optional
        The degree of the fitting polynomial. Only used when interp_method='polynomial'.
        This argument is passed to numpy polyfit `deg` parameter., by default 7.

    Raises
    ------
    ValueError
        Raised if invalid curve or direction argument passed.
    ValueError
        Raised if invalid interp_method argument passed.

    References
    ----------
    .. [1]  Satopaa, V., J., Albrecht, D., Irwin, B., Raghavan. 2011.
            "Finding a "Kneedle" in a Haystack: Detecting Knee Points in System Behavior".
            31st International Conference on Distributed Computing Systems Workshops.
            doi: 10.1109/ICDCSW.2011.20.
    '''
    def __init__(
        self,
        x: numpy.ndarray,
        y: numpy.ndarray,
        S: float = 1.0,
        curve: str = 'concave',
        direction: str = 'increasing',
        interp_method: str = 'interp1d',
        online: bool = False,
        polynomial_degree: int = 7,
    ):
        # Step 0: Raw Input
        self.x = numpy.array(x)
        self.y = numpy.array(y)
        self.curve = curve
        self.direction = direction
        self.N = len(self.x)
        self.S = S
        self.all_knees = set()
        self.all_norm_knees = set()
        self.all_knees_y = []
        self.all_norm_knees_y = []
        self.online = online
        self.polynomial_degree = polynomial_degree

        valid_curve = self.curve in VALID_CURVE
        valid_direction = self.direction in VALID_DIRECTION
        if not all((valid_curve, valid_direction)):
            raise ValueError(
                'Please check that the curve and direction arguments are valid.'
            )

        # Step 1: fit a smooth line
        if interp_method == 'interp1d':
            # uspline = interpolate.interp1d(self.x, self.y)
            # self.ds_y = uspline(self.x)
            self.ds_y = y
        elif interp_method == 'polynomial':
            p = numpy.poly1d(numpy.polyfit(x, y, self.polynomial_degree))
            self.ds_y = p(x)
        else:
            msg = f'{interp_method} is an invalid interp_method parameter, ' +\
                'use either "interp1d" or "polynomial".'
            raise ValueError(msg)

        # Step 2: normalize values
        self.x_normalized = self._normalize(self.x)
        self.y_normalized = self._normalize(self.ds_y)

        # Step 3: Calculate the Difference curve
        self.y_normalized = self.transform_y(
            self.y_normalized, self.direction, self.curve
        )
        # normalized difference curve
        self.y_difference = self.y_normalized - self.x_normalized
        self.x_difference = self.x_normalized.copy()

        # Step 4: Identify local maxima/minima
        # local maxima
        self.maxima_indices = rel_extrema_idcs(self.y_difference, numpy.greater_equal)
        self.x_difference_maxima = self.x_difference[self.maxima_indices]
        self.y_difference_maxima = self.y_difference[self.maxima_indices]

        # local minima
        self.minima_indices = rel_extrema_idcs(self.y_difference, numpy.less_equal)
        self.x_difference_minima = self.x_difference[self.minima_indices]
        self.y_difference_minima = self.y_difference[self.minima_indices]

        # Step 5: Calculate thresholds
        self.Tmx = self.y_difference_maxima - (
            self.S * numpy.abs(numpy.diff(self.x_normalized).mean())
        )

        # Step 6: find knee
        self.knee, self.norm_knee = self.find_knee()

        # Step 7: If we have a knee, extract data about it
        self.knee_y = self.norm_knee_y = None
        if self.knee:
            self.knee_y = self.y[self.x == self.knee][0]
            self.norm_knee_y = self.y_normalized[self.x_normalized == self.norm_knee][0]

    @staticmethod
    def _normalize(x: numpy.ndarray) -> numpy.ndarray:
        '''_normalize

        Scale vector values between 0 and 1.

        Parameters
        ----------
        x : numpy.ndarray
            Vector to scale.

        Returns
        -------
        numpy.ndarray
            Scaled vector
        '''
        return (x - x.min()) / (x.max() - x.min())

    @staticmethod
    def transform_y(y: Iterable[float], direction: str, curve: str) -> float:
        '''transform y to concave, increasing based on given direction and curve'''
        # convert elbows to knees
        if direction == 'decreasing':
            if curve == 'concave':
                y = numpy.flip(y)
            elif curve == 'convex':
                y = y.max() - y
        elif direction == 'increasing' and curve == 'convex':
            y = numpy.flip(y.max() - y)

        return y

    def find_knee(self,):
        '''
        This function is called when KneeLocator is instantiated.
        It identifies the knee value and sets the instance attributes.
        '''
        if not self.maxima_indices.size:
            warnings.warn(
                'No local maxima found in the difference curve\n'
                'The line is probably not polynomial.',
                RuntimeWarning,
            )
            return None, None
        # placeholder for which threshold region i is located in.
        maxima_threshold_index = 0
        minima_threshold_index = 0
        # traverse the difference curve
        for i, x in enumerate(self.x_difference):
            # skip points on the curve before the the first local maxima
            if i < self.maxima_indices[0]:
                continue

            j = i + 1

            # reached the end of the curve
            if x == 1.0:
                break

            # if we're at a local max, increment the maxima threshold index and continue
            if (self.maxima_indices == i).any():
                threshold = self.Tmx[maxima_threshold_index]
                threshold_index = i
                maxima_threshold_index += 1
            # values in difference curve are at or after a local minimum
            if (self.minima_indices == i).any():
                threshold = 0.0
                minima_threshold_index += 1

            if self.y_difference[j] < threshold:
                if self.curve == 'convex':
                    if self.direction == 'decreasing':
                        knee = self.x[threshold_index]
                        norm_knee = self.x_normalized[threshold_index]
                    else:
                        knee = self.x[-(threshold_index + 1)]
                        norm_knee = self.x_normalized[threshold_index]

                elif self.curve == 'concave':
                    if self.direction == 'decreasing':
                        knee = self.x[-(threshold_index + 1)]
                        norm_knee = self.x_normalized[threshold_index]
                    else:
                        knee = self.x[threshold_index]
                        norm_knee = self.x_normalized[threshold_index]

                # add the y value at the knee
                y_at_knee = self.y[self.x == knee][0]
                y_norm_at_knee = self.y_normalized[self.x_normalized == norm_knee][0]
                if knee not in self.all_knees:
                    self.all_knees_y.append(y_at_knee)
                    self.all_norm_knees_y.append(y_norm_at_knee)

                # now add the knee
                self.all_knees.add(knee)
                self.all_norm_knees.add(norm_knee)

                # if detecting in offline mode, return the first knee found
                if self.online is False:
                    return knee, norm_knee

        if self.all_knees == set():
            warnings.warn('No knee/elbow found')
            return None, None

        return knee, norm_knee

    # Niceties for users working with elbows rather than knees
    @property
    def elbow(self):
        return self.knee

    @property
    def norm_elbow(self):
        return self.norm_knee

    @property
    def elbow_y(self):
        return self.knee_y

    @property
    def norm_elbow_y(self):
        return self.norm_knee_y

    @property
    def all_elbows(self):
        return self.all_knees

    @property
    def all_norm_elbows(self):
        return self.all_norm_knees

    @property
    def all_elbows_y(self):
        return self.all_knees_y

    @property
    def all_norm_elbows_y(self):
        return self.all_norm_knees_y



''' fasta

    Module for reading and writing PHYLIP files.
'''

import os
import re
from typing import Tuple, Union

import numpy

import pyckmeans.distance

WHITESPACE_RE = re.compile(r'\s+')

class InvalidPhylipAlignmentError(Exception):
    '''InvalidPhylipAlignmentError
    '''

def read_phylip_alignment(
    phylip_file: str,
    dtype: Union[str, numpy.dtype] = 'U',
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    '''read_phylip_alignment

    Read phylip alignment file. This function expects the phylip to be a valid alignment,
    meaning that it should contain at least 2 sequences of the same length, including
    gaps.

    WARNING: whitespace characters in entry names are NOT supported.

    Parameters
    ----------
    phylip_file : str
        Path to a phylip file.
    dtype: Union[str, numpy.dtype]
        Data type to use for the sequence array.

    Returns
    -------
    Tuple[numpy.ndarray, numpy.ndarray]
        Tuple of sequences and names, each as numpy array.

    Raises
    ------
    InvalidPhylipAlignmentError
        Raised if header is malformed.
    InvalidPhylipAlignmentError
        Raised if less than 2 entries are present in phylip_file.
    InvalidPhylipAlignmentError
        Raised if number of entries does not match header.
    '''

    names = []
    seqs = []
    with open(phylip_file) as phylip_f:
        # header
        header_str = next(phylip_f)
        try:
            n_entries, n_sites = [int(s) for s in header_str.split()]
        except:
            raise InvalidPhylipAlignmentError('Malformed header.')

        for line in phylip_f:
            _line = re.sub(WHITESPACE_RE, '', line)
            if not _line:
                continue
            l_len = len(_line)
            start = l_len-n_sites
            name = _line[:start]
            seq = _line[start:].upper()

            names.append(name)
            seqs.append(list(seq))

    # check alignment validity
    n_seq = len(seqs)
    if len(seqs) < 2:
        msg = f'Expected at least 2 entries but found only {n_seq}.'
        raise InvalidPhylipAlignmentError(msg)

    if n_seq != n_entries:
        msg = f'Expected {n_entries} entries but found {n_seq} instead.'
        raise InvalidPhylipAlignmentError(msg)

    # construct output
    seqs = numpy.array(seqs, dtype=dtype)
    names = numpy.array(names)

    return seqs, names


class InvalidPhylipMatrixError(Exception):
    '''InvalidPhylipMatrixTypeError
    '''

def read_phylip_distmat(phylip_file: str) -> 'pyckmeans.distance.DistanceMatrix':
    '''read_phylip_distmat

    Read distance matrix in PHYLIP format.
    Supports full and lower-triangle matrices.

    Parameters
    ----------
    phylip_file : str
        Path to distance file in phylip format.

    Returns
    -------
    pyckmeans.distance.DistanceMatrix
        Distance matrix as pyckmeans.distance DistanceMatrix object.

    Raises
    ------
    InvalidPhylipMatrixError
        Raised if the header is malformed.
    InvalidPhylipMatrixError
        Raised if an empty line is encountered as second line.
    InvalidPhylipMatrixError
        Raised if file format can neither be inferred as full nor
        as lower-triangle matrix.
    InvalidPhylipMatrixError
        Raised if an empty line is encountered.
    InvalidPhylipMatrixError
        Raised if expecting a full matrix but number of values
        does not match the header.
    InvalidPhylipMatrixError
        Raised if an empty line is encountered.
    InvalidPhylipMatrixError
        Raised if expecting lower-triangle matrix but number of values
        does not match the expected number of values for that entry.
    InvalidPhylipMatrixError
        Raised if number of names does not match number of entries
        stated in the header.
    '''
    with open(phylip_file) as phylip_f:
        # == header
        header_str = next(phylip_f)
        try:
            n_entries = int(header_str.strip())
        except:
            raise InvalidPhylipMatrixError('Malformed header.')

        dist_mat = numpy.zeros((n_entries, n_entries))
        names = []

        # == detect matrix type (full, lower-triangle)
        line = next(phylip_f)
        _line = line.strip()
        if not _line:
            msg = 'Line 2: Empty lines are not allowed.'
            raise InvalidPhylipMatrixError(msg)
        name, *mat_entries = _line.split()
        names.append(name)

        # lower-triangle matrix
        if len(mat_entries) == 0:
            mat_type = 'lower-triangle'
        # full matrix
        elif len(mat_entries) == n_entries:
            mat_type = 'full'
            dist_mat[0,] = numpy.array(mat_entries, dtype=float)
        # error
        else:
            msg = 'Line 2: Expected either 0 values for a lower-triangle ' +\
                f'matrix or {n_entries} values for a full matrix; found ' +\
                f'{len(mat_entries)} values instead.'
            raise InvalidPhylipMatrixError(msg)

        # == full matrix
        if mat_type == 'full':
            for i, line in enumerate(phylip_f):
                l_num = i + 3 # 1-based line number: header + first line already read

                _line = line.strip()
                if not _line:
                    # last line can be empty
                    if i + 2 == n_entries:
                        continue
                    msg = f'Line {l_num}: Empty lines are not allowed.'
                    raise InvalidPhylipMatrixError(msg)
                name, *mat_entries = _line.split()
                names.append(name)

                # error
                if len(mat_entries) != n_entries:
                    msg = f'Line {l_num}: Expected {n_entries} values for a full matrix but ' +\
                        f'found {len(mat_entries)} values instead.'
                    raise InvalidPhylipMatrixError(msg)

                dist_mat[i+1,] = numpy.array(mat_entries, dtype=float)

        # == lower-triangle matrix
        elif mat_type == 'lower-triangle':
            for i, line in enumerate(phylip_f):
                l_num = i + 3 # 1-based line number: header + first line already read

                _line = line.strip()
                if not _line:
                    # last line can be empty
                    if i + 2 == n_entries:
                        continue
                    msg = f'Line {l_num}: Empty lines are not allowed.'
                    raise InvalidPhylipMatrixError(msg)
                name, *mat_entries = _line.split()
                names.append(name)

                # error
                if len(mat_entries) != i+1:
                    msg = f'Line {l_num}: Expected {i+1} values for a lower-triangle ' +\
                        f'matrix but found {len(mat_entries)} values instead.'
                    raise InvalidPhylipMatrixError(msg)

                dist_mat[i+1, :i+1] = numpy.array(mat_entries, dtype=float)

            # fill upper triangle
            dist_mat = dist_mat + dist_mat.T

    # check validity
    if len(names) != n_entries:
        msg = f'Expected {n_entries} entries but found {len(names)}.'
        raise InvalidPhylipMatrixError(msg)

    return pyckmeans.distance.DistanceMatrix(dist_mat, names)

class IncompatibleNamesError(Exception):
    '''IncompatibleNamesError'''

NAME_PADDING = 64

def write_phylip_distmat(
    dist: 'pyckmeans.distance.DistanceMatrix',
    file_path: str,
    force: bool = False,
) -> None:
    '''write_phylip_distmat

    Write distance matrix to file in PHYLIP matrix format.

    Parameters
    ----------
    dist : pyckmeans.distance.DistanceMatrix
        Distance matrix as pyckmeans.distance DistanceMatrix object.
    file_path : str
        Output file path.
    force : bool, optional
        Force overwrite if file exists, by default False

    Raises
    ------
    FileExistsError
        Raised if file at file_path already exists and force is False.
    FileExistsError
        Raised if file_path points to an existing directory.
    IncompatibleNamesError
        Raised if names are incompatible with dist_mat.
    '''
    if os.path.exists(file_path):
        if os.path.isfile(file_path) and not force:
            msg = f'File {file_path} already exists. If you want to overwrite ' +\
                'it run the function with force=True.'
            raise FileExistsError(msg)
        else:
            msg = f'A directory exists at path {file_path}.'
            raise FileExistsError(msg)

    dist_mat = dist.dist_mat
    names = dist.names

    n_entries = dist_mat.shape[0]
    if len(names) != n_entries:
        msg = f'Expected {n_entries} names but got {len(names)} instead.'
        raise IncompatibleNamesError(msg)

    with open(file_path, 'w') as phylip_f:
        # header
        phylip_f.write(f'{n_entries}\n')

        # body
        for name, dists in zip(names, dist_mat):
            nm_str = f'{name: <{NAME_PADDING}}'
            dst_str = '\t'.join(dists.astype(str))
            phylip_f.write(f'{nm_str} {dst_str}\n')



''' io

    Module containing input and output functionality.
'''

from .nucleotide_alignment import \
    NucleotideAlignment, \
    read_alignment, \
    InvalidAlignmentFileExtensionError, \
    InvalidAlignmentFileFormatError
from .phylip import \
    read_phylip_alignment, \
    InvalidPhylipAlignmentError, \
    read_phylip_distmat, \
    InvalidPhylipMatrixError
from .fasta import \
    read_fasta_alignment, \
    InvalidFastaAlignmentError
from .csv import \
    read_csv_distmat



''' Plotting utitlies
'''

from pyckmeans.core import wecr
from typing import Iterable, Optional, Tuple, Union
import numpy
import matplotlib.pyplot as plt
import matplotlib.figure
import matplotlib.colors
import matplotlib.axes
import pyckmeans.core

def plot_ckmeans_result(
    ckm_res: pyckmeans.core.CKmeansResult,
    names: Optional[Iterable[str]] = None,
    order: Optional[Union[str, numpy.ndarray]] = 'GW',
    cmap_cm: Union[str, matplotlib.colors.Colormap] = 'Blues',
    cmap_clbar: Union[str, matplotlib.colors.Colormap] = 'tab20',
    figsize: Tuple[float, float] = (7, 7),
) -> matplotlib.figure.Figure:
    '''plot_ckmeans_result

    Plot pyckmeans result consensus matrix with consensus clusters.

    Parameters
    ----------
    ckm_res : CKmeansResult
        CKmeansResult as returned from CKmeans.predict.
    names : Optional[Iterable[str]]
        Sample names to be plotted.
    order : Optional[Union[str, numpy.ndarray]]
        Sample Plotting order. Either a string, determining the oder method to use
        (see CKmeansResult.order), or a numpy.ndarray giving the sample order,
        or None to apply no reordering.
    cmap_cm : Union[str, matplotlib.colors.Colormap], optional
        Colormap for the consensus matrix, by default 'Blues'
    cmap_clbar : Union[str, matplotlib.colors.Colormap], optional
        Colormap for the cluster bar, by default 'tab20'
    figsize : Tuple[float, float], optional
        Figure size for the matplotlib figure, by default (7, 7).

    Returns
    -------
    matplotlib.figure.Figure
        Matplotlib figure.
    '''
    # if order is None do not reorder
    if order is None:
        order = numpy.arange(ckm_res.cmatrix.shape[0])
    # if order is str use CKMeansResult order
    elif isinstance(order, str):
        order = ckm_res.order(method=order)
    # else order must be numpy.ndarray giving the sample order

    ckm_res = ckm_res.reorder(order=order, in_place=False)
    cl = ckm_res.cl

    # if names is passed use names, else try to get names
    # from ckm_res, else just use samples indices
    if names is None:
        if ckm_res.names is not None:
            nms = ckm_res.names
        else:
            nms = order.astype('str')
    else:
        nms = numpy.array(names)[order]

    # build figure layout
    fig = plt.figure(figsize=figsize)
    ax_cmat = fig.add_axes([0.1, 0.1, 0.8, 0.8])
    ax_clbar = fig.add_axes([0.05, 0.1, 0.05, 0.8])
    ax_cbar = fig.add_axes([0.925, 0.1, 0.025, 0.8])

    # = consensus matrix
    ax_cmat.imshow(ckm_res.cmatrix, cmap=cmap_cm)
    ax_cmat.set_xticks(numpy.arange(len(nms)))
    ax_cmat.set_xticklabels(nms)
    for tick in ax_cmat.get_xticklabels():
        tick.set_rotation(90)
    ax_cmat.set_yticks([])
    ax_cmat.tick_params(left=False)

    # cluster lines
    cl_01 = []
    cl_start = 0
    for i in range(1, len(cl)):
        if cl[i] != cl[cl_start]:
            cl_01.append((cl_start, i))
            cl_start = i
    cl_01.append((cl_start, len(cl)))
    cl_01 = numpy.array(cl_01)

    ax_cmat.hlines(cl_01[:, 0] + 0.5 - 1, -0.5, len(nms) - 0.5, color='white', linewidth=2)
    ax_cmat.vlines(cl_01[:, 0] + 0.5 - 1, -0.5, len(nms) - 0.5, color='white', linewidth=2)

    # = cluster membership bar
    ax_clbar.imshow(ckm_res.cl.reshape(-1, 1), cmap=cmap_clbar)
    ax_clbar.set_xticks([])
    ax_clbar.set_yticks(numpy.arange(len(nms)))
    ax_clbar.set_yticklabels(nms)

    # = color bar
    ax_cbar.set_xticks([])
    ax_cbar.yaxis.tick_right()
    plt.colorbar(plt.cm.ScalarMappable(cmap=cmap_cm), cax=ax_cbar)

    return fig

def plot_multickmeans_metrics(
    mckm_res: pyckmeans.core.MultiCKmeansResult,
    figsize: Tuple[float, float] = (7, 7),
) -> matplotlib.figure.Figure:
    '''plot_multickmeans_metrics

    Plot MultiCKMeansResult metrics.

    Parameters
    ----------
    mckm_res : MultiCKmeansResult
        MultiCKmeansResult object
    figsize : Tuple[float, float], optional
        Figure size for the matplotlib figure, by default (7, 7).

    Returns
    -------
    matplotlib.figure.Figure
        Matplotlib Figure of the metrics plot.
    '''

    fig, axs = plt.subplots(nrows=2, ncols=2, figsize=figsize)

    axs = axs.flatten()
    for ax in axs:
        ax.grid(axis='x')
        ax.set_xticks(mckm_res.ks)
        ax.set_xlabel('k')

    axs[0].plot(mckm_res.ks, mckm_res.bics)
    axs[0].set_title('BIC')
    axs[0].set_ylabel('BIC')

    axs[1].plot(mckm_res.ks, mckm_res.dbs)
    axs[1].set_title('DB')
    axs[1].set_ylabel('DB')

    axs[2].plot(mckm_res.ks, mckm_res.sils)
    axs[2].set_title('SIL')
    axs[2].set_ylabel('SIL')

    axs[3].plot(mckm_res.ks, mckm_res.chs)
    axs[3].set_title('CH')
    axs[3].set_ylabel('CH')

    return fig

def plot_wecr_result(
    wecr_res: pyckmeans.core.WECRResult,
    k: int,
    names: Optional[Iterable[str]] = None,
    order: Optional[Union[str, numpy.ndarray]] = 'GW',
    cmap_cm: Union[str, matplotlib.colors.Colormap] = 'Blues',
    cmap_clbar: Union[str, matplotlib.colors.Colormap] = 'tab20',
    figsize: Tuple[float, float] = (7, 7),
) -> matplotlib.figure.Figure:
    '''plot_wecr_result

    Plot wecr result consensus matrix with consensus clusters.

    Parameters
    ----------
    wecr_res : pyckmeans.core.WECRResult
        WECRResult as returned from pyckmeans.core.WECR.predict.
    k: int
        The number of clusters k to use for plotting.
    names : Optional[Iterable[str]]
        Sample names to be plotted.
    order : Optional[Union[str, numpy.ndarray]]
        Sample Plotting order. Either a string, determining the oder method to use
        (see CKmeansResult.order), or a numpy.ndarray giving the sample order,
        or None to apply no reordering.
    cmap_cm : Union[str, matplotlib.colors.Colormap], optional
        Colormap for the consensus matrix, by default 'Blues'
    cmap_clbar : Union[str, matplotlib.colors.Colormap], optional
        Colormap for the cluster bar, by default 'tab20'
    figsize : Tuple[float, float], optional
        Figure size for the matplotlib figure, by default (7, 7).

    Returns
    -------
    matplotlib.figure.Figure
        Matplotlib figure.

    Raises
    ------
    wecr.InvalidKError
        Raised if an invalid k argument is provided.
    '''
    # if order is None do not reorder
    if order is None:
        order = numpy.arange(wecr_res.cmatrix.shape[0])
    # if order is str use WECRResult order
    elif isinstance(order, str):
        order = wecr_res.order(method=order)
    # else order must be numpy.ndarray giving the sample order

    wecr_res = wecr_res.reorder(order=order, in_place=False)
    if not k in wecr_res.k:
        msg = f'Result for k={k} not found. Available k are {wecr_res.k}.'
        raise wecr.InvalidKError(msg)
    cl = wecr_res.cl[numpy.argmax(wecr_res.k == k)]

    # if names is passed use names, else try to get names
    # from wecr_res, else just use samples indices
    if names is None:
        if wecr_res.names is not None:
            nms = wecr_res.names
        else:
            nms = order.astype('str')
    else:
        nms = numpy.array(names)[order]

    # build figure layout
    fig = plt.figure(figsize=figsize)
    ax_cmat = fig.add_axes([0.1, 0.1, 0.8, 0.8])
    ax_clbar = fig.add_axes([0.05, 0.1, 0.05, 0.8])
    ax_cbar = fig.add_axes([0.925, 0.1, 0.025, 0.8])

    # = consensus matrix
    ax_cmat.imshow(wecr_res.cmatrix, cmap=cmap_cm)
    ax_cmat.set_xticks(numpy.arange(len(nms)))
    ax_cmat.set_xticklabels(nms)
    for tick in ax_cmat.get_xticklabels():
        tick.set_rotation(90)
    ax_cmat.set_yticks([])
    ax_cmat.tick_params(left=False)

    # cluster lines
    cl_01 = []
    cl_start = 0
    for i in range(1, len(cl)):
        if cl[i] != cl[cl_start]:
            cl_01.append((cl_start, i))
            cl_start = i
    cl_01.append((cl_start, len(cl)))
    cl_01 = numpy.array(cl_01)

    ax_cmat.hlines(cl_01[:, 0] + 0.5 - 1, -0.5, len(nms) - 0.5, color='white', linewidth=2)
    ax_cmat.vlines(cl_01[:, 0] + 0.5 - 1, -0.5, len(nms) - 0.5, color='white', linewidth=2)

    # = cluster membership bar
    ax_clbar.imshow(cl.reshape(-1, 1), cmap=cmap_clbar)
    ax_clbar.set_xticks([])
    ax_clbar.set_yticks(numpy.arange(len(nms)))
    ax_clbar.set_yticklabels(nms)

    # = color bar
    ax_cbar.set_xticks([])
    ax_cbar.yaxis.tick_right()
    plt.colorbar(plt.cm.ScalarMappable(cmap=cmap_cm), cax=ax_cbar)

    return fig

def plot_wecr_result_metrics(
    wecr_res: pyckmeans.core.WECRResult,
    figsize: Tuple[float, float] = (7, 7),
) -> matplotlib.figure.Figure:
    '''plot_wecr_result_metrics

    Plot WECRResult metrics.

    Parameters
    ----------
    wecr_res : WECRResult
        WECRResult object
    figsize : Tuple[float, float], optional
        Figure size for the matplotlib figure, by default (7, 7).

    Returns
    -------
    matplotlib.figure.Figure
        Matplotlib Figure of the metrics plot.
    '''

    fig, axs = plt.subplots(nrows=2, ncols=2, figsize=figsize)

    order = numpy.argsort(wecr_res.k)

    axs = axs.flatten()
    for ax in axs:
        ax.grid(axis='x')
        ax.set_xticks(wecr_res.k[order])
        ax.set_xlabel('k')

    axs[0].plot(wecr_res.k[order], wecr_res.bic[order])
    axs[0].set_title('BIC')
    axs[0].set_ylabel('BIC')

    axs[1].plot(wecr_res.k[order], wecr_res.db[order])
    axs[1].set_title('DB')
    axs[1].set_ylabel('DB')

    axs[2].plot(wecr_res.k[order], wecr_res.sil[order])
    axs[2].set_title('SIL')
    axs[2].set_ylabel('SIL')

    axs[3].plot(wecr_res.k[order], wecr_res.ch[order])
    axs[3].set_title('CH')
    axs[3].set_ylabel('CH')

    return fig



'''core utilities'''

import json
import numpy

# source: https://stackoverflow.com/a/49677241
class NumpyEncoder(json.JSONEncoder):
    ''' Special json encoder for numpy types '''
    def default(self, obj):
        if isinstance(obj, numpy.integer):
            return int(obj)
        elif isinstance(obj, numpy.floating):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



''' distance

    Module for distance calculations.
'''

from typing import Iterable, Optional, Tuple

import numpy

import pyckmeans.io

from .c_interop import p_distance, jc_distance, k2p_distance

class IncompatibleNamesError(Exception):
    '''IncompatibleNamesError'''

class DistanceMatrix:
    '''__init__

    Distance Matrix, optionally named.

    Parameters
    ----------
    dist_mat : numpy.ndarray
        n*n distance matrix.
    names : Optional[Iterable[str]]
        Names, by default None.

    Raises
    ------
    IncompatibleNamesError
        Raised if dimension of names and dist_mat are incompatible.
    '''
    def __init__(self, dist_mat: numpy.ndarray, names: Optional[Iterable[str]] = None):
        self.dist_mat = dist_mat
        self.names = None

        if not names is None:
            n = dist_mat.shape[0]
            if len(names) != n:
                msg = f'Expected {n} names for {n}x{n} distance matrix ' +\
                    f'but {len(names)} were passed.'
                raise IncompatibleNamesError(msg)

            self.names = numpy.array(names)

    def __repr__(self) -> str:
        '''__repr__

        Returns
        -------
        str
            String representation.
        '''
        return f'{repr(self.names)}\n{repr(self.dist_mat)}'

    @property
    def shape(self) -> Tuple[int]:
        '''shape

        Get matrix shape.

        Returns
        -------
        Tuple[int]
            Matrix shape.
        '''
        return self.dist_mat.shape

    @staticmethod
    def from_phylip(file_path: str) -> 'DistanceMatrix':
        '''from_phylip

        Read PHYLIP distance matrix.

        Returns
        -------
        DistanceMatrix
            DistanceMatrix object.
        '''
        return pyckmeans.io.phylip.read_phylip_distmat(file_path)

    @staticmethod
    def from_csv( # pylint: disable=missing-param-doc
        file_path: str,
        header: Optional[int] = 0,
        index_col: Optional[int] = 0,
        sep: str = ',',
        **kwargs,
    ) -> 'DistanceMatrix':
        '''read_csv_distmat

        Read distance matrix from CSV file.

        Parameters
        ----------
        file_path : str
            Path to CSV file.
        header : Optional[int]
            Determines the row in the CSV file containing
            sample names. Is passed to pandas.read_csv(). By default 0, meaning
            the first row.
        index_col : Optional[int]
            Determines the index column. By default, the first column is expected
            to contain sample names. Passed to pandas.read_csv().
        sep : str
            Column separator, be default ','. Passed to Passed to pandas.read_csv().
        **kwargs
            Additional keyword arguments passed to pandas.read_csv().
        Returns
        -------
        pyckmeans.distance.DistanceMatrix
            DistanceMatrix object.
        '''
        return pyckmeans.io.csv.read_csv_distmat(
            file_path=file_path,
            header=header,
            index_col=index_col,
            sep=sep,
            **kwargs,
        )

    def to_phylip(
        self,
        file_path: str,
        force: bool = False,
    ):
        '''to_phylip

        Write distance matrix to file in PHYLIP matrix format.

        Parameters
        ----------
        file_path : str
            Output file path.
        force : bool, optional
            Force overwrite if file exists, by default False
        '''
        pyckmeans.io.phylip.write_phylip_distmat(
            dist=self,
            file_path=file_path,
            force=force,
        )

    def to_csv(
        self,
        file_path: str,
        force: bool = False,
    ):
        '''to_csv

        Write DistanceMatrix object to CSV.

        Parameters
        ----------
        file_path : str
            CSV file path.
        force : bool, optional
            Force overwrite if file_path already exists, by default False
        '''
        pyckmeans.io.csv.write_csv_distmat(
            dist=self,
            file_path=file_path,
            force=force,
        )

class InvalidDistanceTypeError(Exception):
    '''UnknownDistanceTypeError'''

def alignment_distance(
    alignment: "pyckmeans.io.NucleotideAlignment",
    distance_type: str = 'p',
    pairwise_deletion: bool = True,
) -> DistanceMatrix:
    '''genetic_distance

    Calculate genetic distance based on a nucleotide alignment.

    Parameters
    ----------
    alignment : pyckmeans.io.NucleotideAlignment
        Nucleotide alignment.
    distance_type : str, optional
        Type of genetic distance to calculate, by default 'p'.
        Available distance types are p-distances ('p'),
        Jukes-Cantor distances ('jc'), and Kimura 2-paramater distances
        ('k2p').
    pairwise_deletion : bool
        Use pairwise deletion as action to deal with missing data.
        If False, complete deletion is applied.
        Gaps ("-", "~", " "), "?", and ambiguous bases are treated as
        missing data.
    Returns
    -------
    DistanceMatrix
        n*n distance matrix.

    Raises
    ------
    InvalidDistanceTypeError
        Raised if invalid distance_type is passed.
    '''
    distance_type = distance_type.lower()
    if distance_type in ['p', 'raw']:
        return DistanceMatrix(
            p_distance(alignment.sequences, pairwise_deletion),
            alignment.names,
        )
    elif distance_type in ['jc', 'jc69']:
        return DistanceMatrix(
            jc_distance(alignment.sequences, pairwise_deletion),
            alignment.names,
        )
    elif distance_type in ['k2p', 'k80']:
        return DistanceMatrix(
            k2p_distance(alignment.sequences, pairwise_deletion),
            alignment.names,
        )
    else:
        msg = f'Unknown distance type "{distance_type}".'
        raise InvalidDistanceTypeError(msg)



import ctypes
import pathlib

import numpy

# load the shared library
libfile = pathlib.Path(__file__).parent / 'lib' / 'distance.so'
lib = ctypes.CDLL(str(libfile))

# == p distance
lib.pDistance.restype = None
lib.pDistance.argtypes = [
    numpy.ctypeslib.ndpointer(  # alignment: n * m matrix
        dtype=numpy.uint8,
        ndim=2,
        flags='C_CONTIGUOUS',
    ),
    ctypes.c_int,               # n: number of entries
    ctypes.c_int,               # m: number of sites
    ctypes.c_int,               # pairwiseDeletion
    numpy.ctypeslib.ndpointer(  # (output) distMat: n * n distance matrixmatrix
        dtype=numpy.double,
        ndim=2,
        flags='C_CONTIGUOUS',
    ),
]

def p_distance(
    alignment: numpy.ndarray,
    pairwise_deletion: bool = True,
) -> numpy.ndarray:
    '''p_distance

    Calculate p-distance for a nucleotide alignment.

    Parameters
    ----------
    alignment : numpy.ndarray
        n*m numpy alignment, where n is the number of entries and m is
        the number of sites. Bases must be encoded in the format of
        pyckmeans.io.NucleotideAlignment.
    pairwise_deletion : bool, optional
        Calculate distances with pairwise-deletion in case of missing
        data, by default True

    Returns
    -------
    numpy.ndarray
        n*n distance matrix.
    '''
    if not alignment.flags['C_CONTIGUOUS']:
        alignment = numpy.ascontiguousarray(alignment)

    n, m = alignment.shape

    dist_mat = numpy.zeros((n, n), dtype=numpy.double)

    lib.pDistance(alignment, n, m, pairwise_deletion, dist_mat)

    return dist_mat

# == Jukes-Cantor distance
lib.jcDistance.restype = None
lib.jcDistance.argtypes = [
    numpy.ctypeslib.ndpointer(  # alignment: n * m matrix
        dtype=numpy.uint8,
        ndim=2,
        flags='C_CONTIGUOUS',
    ),
    ctypes.c_int,               # n: number of entries
    ctypes.c_int,               # m: number of sites
    ctypes.c_int,               # pairwiseDeletion
    numpy.ctypeslib.ndpointer(  # (output) distMat: n * n distance matrixmatrix
        dtype=numpy.double,
        ndim=2,
        flags='C_CONTIGUOUS',
    ),
]

def jc_distance(
    alignment: numpy.ndarray,
    pairwise_deletion: bool = True,
) -> numpy.ndarray:
    '''jc_distance

    Calculate Jukes-Cantor distance for a nucleotide alignment.

    Parameters
    ----------
    alignment : numpy.ndarray
        n*m numpy alignment, where n is the number of entries and m is
        the number of sites. Bases must be encoded in the format of
        pyckmeans.io.NucleotideAlignment.
    pairwise_deletion : bool, optional
        Calculate distances with pairwise-deletion in case of missing
        data, by default True

    Returns
    -------
    numpy.ndarray
        n*n distance matrix.
    '''
    if not alignment.flags['C_CONTIGUOUS']:
        alignment = numpy.ascontiguousarray(alignment)

    n, m = alignment.shape

    dist_mat = numpy.zeros((n, n), dtype=numpy.double)

    lib.jcDistance(alignment, n, m, pairwise_deletion, dist_mat)

    return dist_mat

# == Kimura 2-parameter distance
lib.k2pDistance.restype = None
lib.k2pDistance.argtypes = [
    numpy.ctypeslib.ndpointer(  # alignment: n * m matrix
        dtype=numpy.uint8,
        ndim=2,
        flags='C_CONTIGUOUS',
    ),
    ctypes.c_int,               # n: number of entries
    ctypes.c_int,               # m: number of sites
    ctypes.c_int,               # pairwiseDeletion
    numpy.ctypeslib.ndpointer(  # (output) distMat: n * n distance matrixmatrix
        dtype=numpy.double,
        ndim=2,
        flags='C_CONTIGUOUS',
    ),
]

def k2p_distance(
    alignment: numpy.ndarray,
    pairwise_deletion: bool = True,
) -> numpy.ndarray:
    '''jc_distance

    Calculate Kimura 2-parameter distance for a nucleotide alignment.

    Parameters
    ----------
    alignment : numpy.ndarray
        n*m numpy alignment, where n is the number of entries and m is
        the number of sites. Bases must be encoded in the format of
        pyckmeans.io.NucleotideAlignment.
    pairwise_deletion : bool, optional
        Calculate distances with pairwise-deletion in case of missing
        data, by default True

    Returns
    -------
    numpy.ndarray
        n*n distance matrix.
    '''
    if not alignment.flags['C_CONTIGUOUS']:
        alignment = numpy.ascontiguousarray(alignment)

    n, m = alignment.shape

    dist_mat = numpy.zeros((n, n), dtype=numpy.double)

    lib.k2pDistance(alignment, n, m, pairwise_deletion, dist_mat)

    return dist_mat



import pandas
import pytest
import tempfile
import os

import numpy as np
from sklearn.datasets import make_blobs

from pyckmeans.core.multickmeans import MultiCKMeans
from pyckmeans.core.ckmeans import CKmeans, CKmeansResult, InvalidClusteringMetric
from pyckmeans.core.wecr import WECR, WECRResult, InvalidConstraintsError, InvalidKError

@pytest.fixture(scope='session')
def test_dir():
    with tempfile.TemporaryDirectory() as tempdir:

        yield tempdir

        print(f'Deleted temporary directory {tempdir}.')

def assert_ckm_res_equal(a: CKmeansResult, b: CKmeansResult, eps=1e-8):
    assert (np.abs(a.cmatrix - b.cmatrix) < eps).all()
    assert (a.cl == b.cl).all()
    assert (a.bic is b.bic) or (a.bic - b.bic) < eps
    assert (a.db is b.db) or (a.db - b.db) < eps
    assert (a.sil is b.sil) or (a.sil - b.sil) < eps
    assert (a.ch is b.ch) or (a.ch - b.ch) < eps
    assert (a.names == b.names).all()
    assert a.km_cls is b.km_cls or \
        (np.abs(a.km_cls - b.km_cls) < eps).all()

def assert_wecr_res_equal(a: WECRResult, b: WECRResult, eps=1e-8):
    assert (np.abs(a.cmatrix - b.cmatrix) < eps).all()
    assert (a.cl == b.cl).all()
    assert (a.bic is b.bic) or ((a.bic - b.bic) < eps).all()
    assert (a.db is b.db) or ((a.db - b.db) < eps).all()
    assert (a.sil is b.sil) or ((a.sil - b.sil) < eps).all()
    assert (a.ch is b.ch) or ((a.ch - b.ch) < eps).all()
    assert (a.names == b.names).all()
    assert a.km_cls is b.km_cls or \
        (np.abs(a.km_cls - b.km_cls) < eps).all()

def test_simple():
    ckm_0 = CKmeans(2)
    ckm_1 = CKmeans(np.array(3, dtype=int))
    ckm_2 = CKmeans(np.array(3, dtype=np.int64))

def test_ckmeans():
    x_0, _ = make_blobs(100, 5, centers=3, center_box=[-15, 15], shuffle=False)
    ckm_0 = CKmeans(3, metrics=['sil', 'bic', 'db', 'ch'])
    ckm_0.fit(x_0)
    ckm_res_0 = ckm_0.predict(x_0)

    with pytest.raises(InvalidClusteringMetric):
        CKmeans(3, metrics=['NONEXISTENT_METRIC'])

def test_wecr():
    x_0, _ = make_blobs(100, 5, centers=3, center_box=[-15, 15], shuffle=False)
    wecr_0 = WECR([2,3,4,5], 100)
    wecr_0.fit(x_0)
    
    wecr_res_0 = wecr_0.predict(x_0, must_link=[[0, 1], [0, 2]], must_not_link=[[0, 51], [5, 99]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_0, must_link=[[0, 1], [0, 2]], must_not_link=[[0, 51], [5, 100]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_0, must_link=[[0, 1], [101, 2]], must_not_link=[[0, 51], [5, 99]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_0, must_link=[['a', 'b'], ['c', 'd']], must_not_link=[[0, 51], [5, 99]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_0, must_link=[[0, 1], [0, 2]], must_not_link=[['a', 'b'], ['c', 'd']])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_0, must_link=[[0, 1, 0, 2]], must_not_link=[[0, 51], [5, 99]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_0, must_link=[[0, 1], [0, 2]], must_not_link=[[0, 51, 5, 99]])

    x_1 = pandas.DataFrame(x_0)
    wecr_0.fit(x_1)
    wecr_res_1 = wecr_0.predict(x_1, must_link=[[0, 1], [0, 2]], must_not_link=[[0, 51], [5, 99]])
    wecr_res_1 = wecr_0.predict(x_1, must_link=[['0', '1'], ['0', '2']], must_not_link=[[0, 51], [5, 99]])
    wecr_res_1 = wecr_0.predict(x_1, must_link=[[0, 1], [0, 2]], must_not_link=[['0', '51'], ['5', '99']])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_1, must_link=[[0, 1], [0, 2]], must_not_link=[[0, 51], [5, 100]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_1, must_link=[['a', 'b'], ['c', 'd']], must_not_link=[[0, 51], [5, 99]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_1, must_link=[[0, 'b'], ['c', 'd']], must_not_link=[[0, 51], [5, 99]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_1, must_link=[[0, 1], ['c', 'd']], must_not_link=[[0, 51], [5, 99]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_1, must_link=[[0, 1], [0, 'd']], must_not_link=[[0, 51], [5, 99]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_1, must_link=[[0, 1], [0, 2]], must_not_link=[['a', 'b'], ['c', 'd']])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_1, must_link=[[0, 1], [0, 2]], must_not_link=[[0, 51], [5, 'd']])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_1, must_link=[[0, 1, 0, 2]], must_not_link=[[0, 51], [5, 99]])
    with pytest.raises(InvalidConstraintsError):
        wecr_0.predict(x_1, must_link=[[0, 1], [0, 2]], must_not_link=[[0, 51, 5, 99]])

    wecr_res_1_ro = wecr_res_1.reorder(wecr_res_1.order(linkage_type='single'))
    wecr_res_1_sort = wecr_res_1.sort(linkage_type='single')
    assert_wecr_res_equal(wecr_res_1_ro, wecr_res_1_sort)

    wecr_res_1.plot(2)
    wecr_res_1.plot_metrics()

    wecr_res_1_rcm = wecr_res_1.recalculate_cluster_memberships(x_1, 'average', in_place=True)
    assert wecr_res_1_rcm is wecr_res_1

    wecr_res_1_rcm = wecr_res_1.recalculate_cluster_memberships(x_1, 'average', in_place=False)
    assert not wecr_res_1_rcm is wecr_res_1

    cl = wecr_res_1.get_cl(2, with_names=False)
    cl = wecr_res_1.get_cl(2, with_names=True)
    with pytest.raises(InvalidKError):
        wecr_res_1.get_cl(12500)
    with pytest.raises(InvalidKError):
        wecr_res_1.get_cl(-1)

def test_multickmeans():
    x_0, _ = make_blobs(100, 5, centers=3, center_box=[-15, 15], shuffle=False)
    mckm_0 = MultiCKMeans([2,3,4,5], n_rep=25)
    mckm_0.fit(x_0)
    mckm_res_0 = mckm_0.predict(x_0)
    mckm_res_0_ro = mckm_res_0.reorder(mckm_res_0.order(2), in_place=True)
    assert mckm_res_0 is mckm_res_0_ro

    mckm_res_0_ro = mckm_res_0.reorder(mckm_res_0.order(2), in_place=False)
    assert not mckm_res_0 is mckm_res_0_ro

    with pytest.raises(InvalidClusteringMetric):
        MultiCKMeans([2,3,4,5], n_rep=25, metrics=['NONEXISTENT_METRIC'])

@pytest.mark.parametrize('return_cls', [True, False])
def test_save_load_ckm_res(test_dir, return_cls):
    x, _ = make_blobs(100, 5, centers=3)
    ckm = CKmeans(3, 20)
    ckm.fit(x)
    ckm_res = ckm.predict(x, return_cls=return_cls)

    if return_cls:
        ckm_res_km_cls_file = os.path.join(test_dir, f'{return_cls}_ckm_res_km_cls.txt')
        ckm_res.save_km_cls(ckm_res_km_cls_file, one_hot=False)
        ckm_res.save_km_cls(ckm_res_km_cls_file, one_hot=True)

    assert_ckm_res_equal(
        ckm_res, CKmeansResult.from_json_str(ckm_res.to_json())
    )

    ckm_res_json_file = os.path.join(test_dir, f'{return_cls}_ckm_res.json')
    ckm_res.to_json(ckm_res_json_file)
    ckm_res_l = CKmeansResult.from_json(ckm_res_json_file)

    assert_ckm_res_equal(ckm_res, ckm_res_l)

    ckm_res_dir = os.path.join(test_dir, f'{return_cls}_ckm_res')
    ckm_res.to_dir(ckm_res_dir)
    ckm_res.to_dir(ckm_res_dir, force=True)
    with pytest.raises(Exception):
        ckm_res.to_dir(ckm_res_dir, force=False)
    ckm_res_l = CKmeansResult.from_dir(ckm_res_dir)
    assert_ckm_res_equal(ckm_res, ckm_res_l)

    with pytest.raises(Exception):
        CKmeansResult.from_dir('SOME_NONEXISTENT_DIR')

@pytest.mark.parametrize('return_cls', [True, False])
def test_save_load_wecr_res(test_dir, return_cls):
    x, _ = make_blobs(100, 5, centers=3, center_box=[-15, 15])
    wecr = WECR([2,3,4,5], 100)
    wecr.fit(x)
    wecr_res = wecr.predict(x, return_cls=return_cls)

    if return_cls:
        wecr_res_km_cls_file = os.path.join(test_dir, f'{return_cls}_wecr_res_km_cls.txt')
        wecr_res.save_km_cls(wecr_res_km_cls_file, one_hot=False)
        wecr_res.save_km_cls(wecr_res_km_cls_file, one_hot=True)

    assert_wecr_res_equal(
        wecr_res, WECRResult.from_json_str(wecr_res.to_json())
    )

    wecr_res_json_file = os.path.join(test_dir, f'{return_cls}_wecr_res.json')
    wecr_res.to_json(wecr_res_json_file)
    wecr_res_l = WECRResult.from_json(wecr_res_json_file)
    assert_wecr_res_equal(wecr_res, wecr_res_l)

    wecr_res_dir = os.path.join(test_dir, f'{return_cls}_wecr_res')
    wecr_res.to_dir(wecr_res_dir)
    wecr_res.to_dir(wecr_res_dir, force=True)
    with pytest.raises(Exception):
        wecr_res.to_dir(wecr_res_dir, force=False)
    wecr_res_l = WECRResult.from_dir(wecr_res_dir)
    assert_wecr_res_equal(wecr_res, wecr_res_l)

    with pytest.raises(Exception):
        WECRResult.from_dir('SOME_NONEXISTENT_DIR')



''' setup
'''

import re
import io
from distutils.command.build_ext import build_ext as build_ext_orig
from setuptools import setup, find_packages, Extension

# source: https://stackoverflow.com/a/39671214
__version__ = re.search(
    r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
    io.open('pyckmeans/__init__.py', encoding='utf_8_sig').read()
).group(1)

# ==== ctypes extensions
class CTypesExtension(Extension):
    '''CTypesExtension'''

class build_ext(build_ext_orig):
    '''build_ext'''
    def build_extension(self, ext):
        self._ctypes = isinstance(ext, CTypesExtension)
        return super().build_extension(ext)

    def get_export_symbols(self, ext):
        if self._ctypes:
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        if self._ctypes:
            return ext_name + '.so'
        return super().get_ext_filename(ext_name)

distance_module = CTypesExtension(
    'pyckmeans.distance.lib.distance',
    sources=['pyckmeans/distance/src/distance.cpp'],
    language='c++',
)

nucencode_module = CTypesExtension(
    'pyckmeans.io.lib.nucencode',
    sources=['pyckmeans/io/src/nucencode.cpp'],
    language='c++',
)

ext_modules = [
    distance_module,
    nucencode_module,
]

install_requires = [
    'numpy',
    'pandas',
    'scipy',
    'scikit-learn',
    'matplotlib',
    'tqdm',
]

# ====
description = 'A consensus K-Means implementation.'

long_description_content_type = 'text/x-rst'
long_description = 'A consensus K-Means implementation.'
# ====
setup(
    name='pyckmeans',
    version=__version__,
    packages=find_packages(),
    description=description,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    author='Tankred Ott',
    platforms=['any'],
    python_requires='>=3.6',
    install_requires=install_requires,
    cmdclass={'build_ext': build_ext},
    ext_modules=ext_modules,
    url='https://github.com/TankredO/pyckmeans',
)



''' ordination

    Module for Principle Coordinate Analysis (PCoA) calculation.
'''

# Inspired by https://github.com/cran/ape/blob/master/R/pcoa.R and
# https://github.com/biocore/scikit-bio/blob/0.5.4/skbio/stats/ordination/_principal_coordinate_analysis.py#L23

from types import coroutine
import typing
from warnings import warn
from typing import Iterable, Optional, Union, Dict, Any
import json
import os

import numpy
import pandas

import pyckmeans.distance
from .utils import NumpyEncoder

class InvalidPCOAResultError(Exception):
    '''InvalidPCOAResultError

    Error, signalling an invalid PCOAResult.
    '''

class IncompatibleNamesError(Exception):
    '''IncompatibleNamesError

    Error, signalling that provided names are incompatible with provided data.
    '''

class InvalidFilterError(Exception):
    '''InvalidFilterError'''

class InvalidOutFormatError(Exception):
    '''InvalidOutFormatError'''

class PCOAResult:
    '''PCOAResult

    Object containing the results of a Principle Coordinate Analysis.

    Parameters
    ----------
    vectors : numpy.ndarray
        n * m matrix of potentially corrected eigenvectors, where
        n is the number of samples and m is the number of retained axes.
    eigvals : numpy.ndarray
        Vector of eigenvalues.
    eigvals_rel : numpy.ndarray
        Vector of relative eigenvalues.
    trace : float
        Trace.
    negative_eigvals : bool
        Bool, determining whether negative eigenvalues are present.
    correction : Optional[str], optional
        Type of correction, by default None
    eigvals_corr_rel : Optional[numpy.ndarray], optional
        Vector of corrected relative eigenvalues, by default None
    trace_corr : Optional[float], optional
        Corrected trace, by default None
    names : Optional[Iterable[str]], optional
        Vector of names, by default None

    Raises
    ------
    InvalidPCOAResultError
        Raised if invalid parameter combinations are provided.
    IncompatibleNamesError
        Raised if provided names are incompatible with provided data.
    '''
    def __init__(
        self,
        vectors: numpy.ndarray,
        eigvals: numpy.ndarray,
        eigvals_rel: numpy.ndarray,
        trace: float,
        negative_eigvals: bool,
        correction: Optional[str] = None,
        eigvals_corr_rel: Optional[numpy.ndarray] = None,
        trace_corr: Optional[float] = None,
        names: Optional[Iterable[str]] = None,
    ):
        self.vectors = vectors
        self.eigvals = eigvals
        self.eigvals_rel = eigvals_rel
        self.eigvals_rel_cum = numpy.cumsum(eigvals_rel)
        self.trace = trace

        self.negative_eigvals = negative_eigvals
        self.correction = correction

        self.eigvals_corr_rel = None
        self.eigvals_corr_rel_cum = None
        self.trace_corr = None

        self.names = None
        if not names is None:
            n = vectors.shape[0]
            if len(names) != n:
                msg = f'Expected {n} names for PCOAResult ' +\
                    f'but {len(names)} were passed.'
                raise IncompatibleNamesError(msg)

            self.names = numpy.array(names)

        if negative_eigvals:
            if eigvals_corr_rel is None:
                msg = 'negative_eigvals is True, expecting eigvals_corr_rel.'
                raise InvalidPCOAResultError(msg)

            self.eigvals_corr_rel = eigvals_corr_rel
            self.eigvals_corr_rel_cum = numpy.cumsum(eigvals_corr_rel)

            if correction:
                if trace_corr is None:
                    msg = 'negative eigenvalue correction is set, ' +\
                        'expecting trace_corr.'
                self.trace_corr = trace_corr

            self.values = pandas.DataFrame({
                'eigvals': self.eigvals,
                'eigvals_rel': self.eigvals_rel,
                'eigvals_rel_cum': self.eigvals_corr_rel_cum,
                'eigvals_rel_corrected': self.eigvals_corr_rel,
                'eigvals_rel_corrected_cum': self.eigvals_corr_rel_cum,
            })
        else:
            self.values = pandas.DataFrame({
                'eigvals': self.eigvals,
                'eigvals_rel': self.eigvals_rel,
                'eigvals_rel_cum': self.eigvals_rel_cum,
            })

    def __repr__(self) -> str:
        str_repr = f'<PCOAResult; neg. eigvals: {self.negative_eigvals}, ' +\
            f'correction: {self.correction}>'

        return str_repr

    def to_dict(
        self,
    ) -> Dict:
        '''to_dict

        Convert PCOAResult to dictionary.

        Returns
        -------
        Dict
            PCOAResult as dictionary.
        '''
        return {
            'vectors': self.vectors,
            'negative_eigvals': self.negative_eigvals,
            'correction': self.correction,
            'eigvals': self.eigvals,
            'eigvals_rel': self.eigvals_rel,
            'eigvals_rel_cum': self.eigvals_rel_cum,
            'eigvals_corr_rel': self.eigvals_corr_rel,
            'eigvals_corr_rel_cum': self.eigvals_corr_rel_cum,
            'trace': self.trace,
            'trace_corr': self.trace_corr,
            'names': self.names,
        }

    @classmethod
    def from_dict(
        cls,
        pcoa_res_dict: Dict,
    ) -> 'PCOAResult':
        '''from_dict

        Construct PCOAResult from dictionary.

        Parameters
        ----------
        pcoa_res_dict : Dict
            PCOAResult as dictionary.

        Returns
        -------
        PCOAResult
            PCOAResult
        '''
        return cls(
            vectors=pcoa_res_dict['vectors'],
            eigvals=pcoa_res_dict['eigvals'],
            eigvals_rel=pcoa_res_dict['eigvals_rel'],
            trace=pcoa_res_dict['trace'],
            negative_eigvals=pcoa_res_dict['negative_eigvals'],
            correction=pcoa_res_dict['correction'],
            eigvals_corr_rel=pcoa_res_dict['eigvals_corr_rel'],
            trace_corr=pcoa_res_dict['trace_corr'],
            names=pcoa_res_dict['names'],
        )

    def to_json(
        self,
        file: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> Optional[str]:
        '''to_json

        Convert PCOAResult to JSON string or file.

        Parameters
        ----------
        file : Optional[str], optional
            File path to write the PCOAResult to or None.
            If None, the JSON string will be returned.
        kwargs : Dict[str, Any]
            Additional keyword arguments passed to json.dump or json.dumps.

        Returns
        -------
        Optional[str]
            None or JSON string.
        '''
        pcoa_res_dict = self.to_dict()
        if file is None:
            return json.dumps(pcoa_res_dict, cls=NumpyEncoder, **kwargs)
        else:
            with open(file, 'w') as json_f:
                json.dump(pcoa_res_dict, json_f, cls=NumpyEncoder, **kwargs)

        return None

    @classmethod
    def from_json_str(
        cls,
        json_str: str,
        **kwargs: Dict[str, Any],
    ) -> 'PCOAResult':
        '''from_json_str

        Construct PCOAResult from JSON string.

        Parameters
        ----------
        json_str: str
            JSON string.
        kwargs : Dict[str, Any]
            Additional keyword arguments passed to json.loads.

        Returns
        -------
        PCOAResult
            PCOAResult
        '''
        json_dict = json.loads(json_str, **kwargs)

        json_dict['vectors'] = numpy.array(json_dict['vectors'])
        json_dict['eigvals'] = numpy.array(json_dict['eigvals'])
        json_dict['eigvals_rel'] = numpy.array(json_dict['eigvals_rel'])
        json_dict['eigvals_corr_rel'] = numpy.array(json_dict['eigvals_corr_rel']) \
            if not numpy.array(json_dict['eigvals_corr_rel']) is None else None
        json_dict['names'] = numpy.array(json_dict['names']) \
            if not json_dict['names'] is None else None

        return cls.from_dict(json_dict)

    @classmethod
    def from_json(
        cls,
        file: str,
        **kwargs: Dict[str, Any],
    ) -> 'PCOAResult':
        '''from_json

        Construct PCOAResult from JSON file.

        Parameters
        ----------
        file : str
            JSON file
        kwargs : Dict[str, Any]
            Additional keyword arguments passed to json.loads.
        Returns
        -------
        PCOAResult
            PCOAResult
        '''
        with open(file, 'r') as json_f:
            json_string = json_f.read()

        return cls.from_json_str(json_string, **kwargs)

    def to_dir(
        self,
        out_dir: str,
        force: bool = False,
    ):
        '''to_dir

        Save PCOAResult to directory.
        The directory will contain the three files 'vectors.csv', comprising the eigenvectors,
        'values.csv', comprising the eigenvalues, and 'others.csv', comprising trace and correction
        data.

        Parameters
        ----------
        out_dir : str
            Output directory. Will be created if it does not exist.
        force : bool, optional
            Write into out_dir even if it does already exist, by default False.

        Raises
        ------
        Exception
            Raised if there is a problem with out_dir.
        '''
        if os.path.exists(out_dir):
            if not force:
                msg = f'Output directory "{out_dir}" already exists.'
                raise Exception(msg)
        else:
            os.mkdir(out_dir)

        vectors_file = os.path.join(out_dir, 'vectors.csv')
        vectors_df = pandas.DataFrame(
            self.vectors,
            index=self.names,
            columns=[f'PCo{n}' for n in range(self.vectors.shape[1])]
        )
        vectors_df.index.name = 'sample'
        vectors_df.to_csv(vectors_file, index=True, header=True)

        values_file = os.path.join(out_dir, 'values.csv')
        self.values.to_csv(values_file, index=True, header=True)

        others_files = os.path.join(out_dir, 'others.csv')
        pandas.DataFrame(
            dict(
                trace=[self.trace],
                trace_corr=[self.trace_corr],
                negative_eigvals=[self.negative_eigvals],
                correction=[self.correction],
            ),
        ).to_csv(others_files, index=False, header=True)

    @classmethod
    def from_dir(
        cls,
        directory: str,
    ) -> 'PCOAResult':
        '''from_dir

        Construct PCOAResult from a directory contraining the three files 'vectors.csv',
        'values.csv', and 'others.csv'. See :func:`<pyckmeans.ordination.PCOAResult.to_dir>`.

        Parameters
        ----------
        directory : str
            PCOAResult directory.

        Returns
        -------
        PCOAResult
            PCOAResult

        Raises
        ------
        Exception
            Raised if there is a problem with directory.
        '''
        if not os.path.exists(directory) or not os.path.isdir(directory):
            msg = f'Could not find directory at "{directory}".'
            raise Exception(msg)

        vectors_file = os.path.join(directory, 'vectors.csv')
        vectors_df = pandas.read_csv(vectors_file, header=0, index_col=0)

        values_file = os.path.join(directory, 'values.csv')
        values = pandas.read_csv(values_file, header=0, index_col=0)

        others_files = os.path.join(directory, 'others.csv')
        others = pandas.read_csv(others_files, header=0, index_col=None)

        trace_corr = others['trace_corr'].values[0]
        trace_corr = trace_corr if not numpy.isnan(trace_corr) else None

        correction = others['correction'].values[0]
        correction = correction if isinstance(correction, str) else None

        return cls(
            vectors=vectors_df.values,
            eigvals=values['eigvals'],
            eigvals_rel=values['eigvals_rel'],
            eigvals_corr_rel=\
                values['eigvals_rel_corrected'] if 'eigvals_rel_corrected' in values.columns else None, # pylint: disable=unsubscriptable-object
            trace=others['trace'].values[0],
            trace_corr=trace_corr,
            negative_eigvals=others['negative_eigvals'].values[0],
            correction=correction,
            names=numpy.array(vectors_df.index),
        )

    def get_vectors(
        self,
        filter_by: Optional[str] = None,
        filter_th: Optional[float] = None,
        out_format: str = 'numpy',
    ) -> Union[numpy.ndarray, pandas.DataFrame]:
        '''getVectors

        Get eigenvectors, potentially filtered by cumulative eigenvalue scores.

        Parameters
        ----------
        filter_by : Optional[str], optional
            Cumulative eigenvalue score to filter by or None if no filtering should be applied,
            by default None.
            Can be one of:

            * 'eigvals_rel_cum'
            * 'eigvals_rel_corrected_cum'

            If filter_by is provided filter_th is required.

        filter_th : Optional[float], optional
            Filtering treshold, by default None.
            Eigenvectors will be retained until cumulative eigenvalue score
            is greater than or equal to filter_th.
        out_format : str, optional
            Output format, by default 'numpy'.
            Must be 'numpy' or 'np' for output as numpy.ndarray (Note: name information
            will be lost), or 'pandas' or 'pd' for output as pandas.DataFrame.

        Returns
        -------
        Union[numpy.ndarray, pandas.DataFrame]
            Eigenvectors.

        Raises
        ------
        InvalidFilterError
            Raised if filter arguments are invalid.
        InvalidOutFormatError
            Raised if an invalid out_format is provided.
        '''
        available_filters = ('eigvals_rel_cum', 'eigvals_rel_corrected_cum')
        output_formats = ('numpy', 'np', 'pandas', 'pd')

        x = self.vectors

        if (not filter_by is None) or (not filter_th is None):
            if filter_th is None:
                raise InvalidFilterError('filter_th is required since filter_by is provided')
            if filter_by is None:
                raise InvalidFilterError('filter_by is required since filter_th is provided')

            if filter_by not in available_filters:
                msg = f'Invalid filter_by argument "{filter_by}". ' +\
                    f'filter_by must be one of {available_filters}.'
                raise InvalidFilterError(msg)

            idcs = (self.values[filter_by] < filter_th)[:x.shape[1]]
            x = x[:, idcs]

        if out_format not in output_formats:
            msg = f'Invalid out_format argument "{out_format}". ' +\
                f'out_format must be one of {output_formats}.'
            raise InvalidOutFormatError(msg)

        if out_format in ('pandas', 'pd'):
            x = pandas.DataFrame(x, index=self.names)

        return x

def _center_mat(dmat: numpy.ndarray) -> numpy.ndarray:
    '''_center_mat

    Center n*n matrix.

    Parameters
    ----------
    dmat : numpy.ndarray
        n*n matrix.

    Returns
    -------
    numpy.ndarray
        Centered matrix.
    '''

    n = dmat.shape[0]
    mat = numpy.full((n, n), -1/n)
    mat[numpy.diag_indices(n)] += 1

    return mat.dot(dmat).dot(mat)

class InvalidCorrectionTypeError(Exception):
    '''InvalidCorrectionTypeError'''

class NegativeEigenvaluesCorrectionError(Exception):
    '''FailedCorrectionError

    Error, signalling that the correction of negative eigenvalues failed.
    '''

class NegativeEigenvaluesWarning(Warning):
    '''NegativeEigenvaluesWarning

    Warning, signalling that negative eigenvalues were encountered.
    '''

def pcoa(
    dist: Union[numpy.ndarray, pyckmeans.distance.DistanceMatrix],
    correction: Optional[str] = None,
    eps: float = 1e-8,
) -> PCOAResult:
    '''pcoa

    Principle Coordinate Analysis.

    Parameters
    ----------
    dist : Union[numpy.ndarray, pyckmeans.distance.DistanceMatrix]
        n*n distance matrix either as numpy ndarray or as pyckmeans DistanceMatrix.
    correction: Optional[str]
        Correction for negative eigenvalues, by default None.
        Available corrections are:
            - None: negative eigenvalues are set to 0
            - lingoes: Lingoes correction
            - cailliez: Cailliet correction
    eps : float, optional
        Eigenvalues smaller than eps will be dropped. By default 0.0001

    Returns
    -------
    PCOAResult
        PCoA result object.

    Raises
    ------
    InvalidCorrectionTypeError
        Raised if an unknown correction type is passed.
    NegativeEigenvaluesCorrectionError
        Raised if correction parameter is set and correction of negative
        eigenvalues is not successful.
    '''
    names = None
    if isinstance(dist, pyckmeans.distance.DistanceMatrix):
        names = dist.names
        x = dist.dist_mat
    else:
        x = dist

    n = x.shape[0]

    if not correction is None and not correction in ['lingoes', 'cailliez']:
        msg = f'Unknown correction type "{correction}". ' +\
            'Available correction types are: "lingoes", "cailliez"'
        raise InvalidCorrectionTypeError(msg)

    # center matrix
    dmat_centered = _center_mat((x * x) / -2)
    trace = numpy.diag(dmat_centered).sum()

    # eigen decomposition
    eigvals, eigvecs = numpy.linalg.eigh(dmat_centered, 'U')

    # order descending
    ord_idcs = numpy.argsort(eigvals)[::-1]
    eigvals = eigvals[ord_idcs]
    eigvecs = eigvecs[:, ord_idcs]

    # get min eigenvalue
    min_eigval = numpy.min(eigvals)

    # set small eigenvalues to 0
    zero_eigval_idcs = numpy.nonzero(numpy.abs(eigvals) < eps)[0]
    eigvals[zero_eigval_idcs] = 0

    # no negative eigenvalues
    if min_eigval > -eps:

        eigvals_rel = eigvals / trace

        fze_idx = len(numpy.nonzero(eigvals > eps)[0]) # index of first zero in eigvals
        vectors = eigvecs[:, :fze_idx] * numpy.sqrt(eigvals[:fze_idx])

        return PCOAResult(
            vectors=vectors,
            eigvals=eigvals,
            eigvals_rel=eigvals_rel,
            trace=trace,
            negative_eigvals=False,
            names=names,
        )

    # negative eigenvalues
    else:
        eigvals_rel = eigvals / trace
        eigvals_rel_corrected = (eigvals - min_eigval) / (trace - (n - 1) * min_eigval)
        if len(zero_eigval_idcs) > 0:
            eigvals_rel_corrected = numpy.r_[
                numpy.delete(eigvals_rel_corrected, zero_eigval_idcs[0]), 0
            ]

        fze_idx = len(numpy.nonzero(eigvals > eps)[0]) # index of first zero in eigvals
        vectors = eigvecs[:, :fze_idx] * numpy.sqrt(eigvals[:fze_idx])

        # negative eigenvalues, no correction
        if not correction:
            warn(
                'Negative eigenvalues encountered but no correction applied. '
                'Negative eigenvalues will be treated as 0.',
                NegativeEigenvaluesWarning,
            )

            return PCOAResult(
                vectors=vectors,
                eigvals=eigvals,
                eigvals_rel=eigvals_rel,
                trace=trace,
                negative_eigvals=True,
                correction=None,
                eigvals_corr_rel=eigvals_rel_corrected,
                names=names,
            )

        # negative eigenvalues, correction

        # -- correct distance matrix
        # lingoes correction
        if correction == 'lingoes':
            corr_1 = -min_eigval

            # corrected distance matrix
            x_ncorr = -0.5 * ((x * x) + 2 * corr_1)
        elif correction == 'cailliez':
            dmat_centered_2 = _center_mat(-0.5 * x)

            # prepare matrix for correction
            upper = numpy.c_[numpy.zeros((x.shape[0], x.shape[0])), 2 * dmat_centered]
            lower = numpy.c_[numpy.diag(numpy.full(x.shape[0], -1)), -4 * dmat_centered_2]
            sp_mat = numpy.r_[upper, lower]

            corr_2 = numpy.max(numpy.real(numpy.linalg.eigvals(sp_mat)))

            # corrected distance matrix
            x_ncorr = -0.5 * (x + corr_2)**2

        # -- apply PCoA to corrected distance matrix
        x_ncorr[numpy.diag_indices(x_ncorr.shape[0])] = 0
        x_ncorr = _center_mat(x_ncorr)
        trace_ncorr = numpy.diag(x_ncorr).sum()

        eigvals_ncorr, eigvecs_ncorr = numpy.linalg.eigh(x_ncorr, 'U')

        # order descending
        ord_idcs_ncorr = numpy.argsort(eigvals_ncorr)[::-1]
        eigvals_ncorr = eigvals_ncorr[ord_idcs_ncorr]
        eigvecs_ncorr = eigvecs_ncorr[:, ord_idcs_ncorr]

        # get min eigenvalue
        min_eigval_ncorr = numpy.min(eigvals_ncorr)

        # set small eigenvalues to 0
        zero_eigval_idcs_ncorr = numpy.nonzero(numpy.abs(eigvals_ncorr) < eps)[0]
        eigvals_ncorr[zero_eigval_idcs_ncorr] = 0

        if min_eigval_ncorr < -eps:
            msg = 'Correction failed. There are still negative eigenvalues after applying ' +\
                f'{correction.capitalize()} correction.'
            raise NegativeEigenvaluesCorrectionError(msg)

        eigvals_ncorr_rel = eigvals_ncorr / trace_ncorr

        fze_idx_ncorr = len(numpy.nonzero(eigvals_ncorr > eps)[0]) # index of first zero in eigvals
        vectors_ncorr = eigvecs_ncorr[:, :fze_idx_ncorr] * numpy.sqrt(eigvals_ncorr[:fze_idx_ncorr])

        return PCOAResult(
            vectors=vectors_ncorr,
            eigvals=eigvals,
            eigvals_rel=eigvals_rel,
            trace=trace,
            negative_eigvals=True,
            correction=correction,
            eigvals_corr_rel=eigvals_ncorr_rel,
            trace_corr=trace_ncorr,
            names=names,
        )
